/*
 *    BuildDt.java
 *    Copyright (C) 1999 Pierre Geurts
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

import java.util.*;
import java.net.*;

/**
 * examples:
 *   java buildDt database/omib.db security 700
 *   java buildDt database/iris.db class 100
 *   java buildDt database/color.db color 1500
 */
public class BuildDt {
  public static void main(String args[]){
    if (args.length<3) {
      System.out.println("usage : database goal learning_set_size [entthres stopthres]");
      return;
    }
    /* creation of a database */
    Database db = new Database();
    
    System.out.println("Database loading:");
    try {
      db.loadDb(new URL(args[0]));
    } catch (Exception e) {
      System.out.println("database file not found");
      return;
    }
    if (db.objects.size==0) {
      System.out.println("no object");
      return;
    }
    ObjectSet ls=db.objects.selectRandom(Integer.parseInt(args[2]));
    ObjectSet ts=db.objects.selectNotIn(ls);

    // select the goal attribute
    Attribute goal=db.findAttribute(args[1]);
    if (goal==null) {
      System.out.println("goal attribute not found");
      return;
    }
    
    // select all candidate attributes
    Vector ca=db.getCandidateAttributes(args[1]);
    if (ca.size()==0) {
      System.out.println("no candidate attributes");
      return;
    }

    double entthres=0.0;
    double stopthres=2.0;
    if (args.length>=4) {
      entthres=(new Double(args[3])).doubleValue();
    }
    if (args.length>=5) {
      stopthres=(new Double(args[4])).doubleValue();
    }
    DecisionTree dt=new DecisionTree(db, ls, ca, (SymbAtt)goal,entthres,stopthres);
    dt.buildDt();
    dt.printDt();
    System.out.println("ls errors : " + 100.0*dt.testDt(ls) + "%");
    System.out.println("ts errors : " + 100.0*dt.testDt(ts) + "%");
  }
}




