/*
 *    DecisionTreeApplet.java
 *    Copyright (C) 1999 Pierre Geurts
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

import java.applet.Applet;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.net.*;
import java.io.*;

/**
 * Displays the tree building process
 */
public class DecisionTreeApplet extends Applet 
implements ActionListener,ItemListener, Runnable
{
  /* Following are workarounds for font size calculations, as Netscape
   * reports unappealing font height/ascents on varying platforms. */
  static final int FIXED_FONT_HEIGHT = 10; 
  static final int FIXED_FONT_ASCENT = 3; 

  boolean        running;
  Thread         t;
  WNode          root;
  WalTreeCanvas  tc;
  SimpleNodeDrawer snd;
  DetailedNodeDrawer dnd;

  /* decision tree data */
  Database db;
  DecisionTree   CurrentDt;
  DtNode CurrentDtn;
  ObjectSet ls;
  ObjectSet ts;

  Button btn_scan;    // scan the attribute 
  Button btn_split;   // split one node (if it is still possible)
  Button btn_build;   // build the tree
  Button btn_refresh; // refresh the graph (actually not useful)
  Label lbl_zoom=new Label("Zoom", Label.RIGHT);
  Button btn_zoominc;
  Button btn_zoomdec;
  Choice AttChoice;   // selection list for the best attribute
  String messagechoice="Click 'Scan Attributes' to see the choices";

  TextField tf_opennode=new TextField(); // description of the current node

  // input parameters
  Panel top=new Panel();
  Label lbl_database=new Label("Database url : ", Label.RIGHT); 
  TextField tf_database= new TextField(65); // to enter the file name
  Button btn_loaddb=new Button("Load Database");
         // to restart with a new database
  Label lbl_goal=new Label("Goal Attribute : ", Label.RIGHT); // goal attribute
  Choice GoalAttributeChoice; // a choice list for the goal attribute (only the symbolic)
                        // (at the same time, it displays all the attribute)
  Label lbl_candatt=new Label("Candidate attributes : ");
  java.awt.List lst_candatt; // a selection list for the candidate attributes
  
  // learning set selection
  Label lbl_learningset=new Label("Learning set :", Label.RIGHT);
  Label lbl_lssize=new Label("size :", Label.RIGHT);
  TextField tf_lssize=new TextField(5);
  Label lbl_lsmode=new Label("Selection mode:", Label.RIGHT);
  Choice SelectionChoiceLs;
  
  // tree parameters
  Label lbl_entthres=new Label("Entropy Threshold : ", Label.RIGHT);
  TextField tf_entthres=new TextField(4);
  Label lbl_scorethres=new Label("Inf X LS size Threshold : ", Label.RIGHT);
  TextField tf_scorethres= new TextField(4);
  double default_scorethres=0.0;
  double default_entthres=0.0;
  
  // start tree button
  Button btn_starttree=new Button("New Tree");
  
  // test parameters
  Label lbl_testset=new Label("Test set :", Label.RIGHT);
  Label lbl_tssize=new Label("size :", Label.RIGHT);
  TextField tf_tssize=new TextField(5);
  Label lbl_tsmode=new Label("Selection mode:", Label.RIGHT);
  Choice SelectionChoiceTs;
  TextField tf_testresult=new TextField(20);
  Button btn_testtree;
  Button btn_crossvalidation;
  
  // panels
  Panel t_p0=new Panel(new GridLayout(3,1,0,0));
  Panel t_loaddb=new Panel(new FlowLayout(FlowLayout.CENTER,5,0));
  Panel t_p1=new Panel(new FlowLayout(FlowLayout.CENTER,5,0));
  Panel t_p2=new Panel(new FlowLayout(FlowLayout.CENTER,5,0));
  Panel t_p3=new Panel(new FlowLayout(FlowLayout.CENTER,5,0));
  Panel t_p4=new Panel(new FlowLayout(FlowLayout.CENTER,5,0));
  Panel t_p5=new Panel(new BorderLayout());
  
  Panel b_p1=new Panel(new FlowLayout(FlowLayout.CENTER,5,0));
  Panel b_p2=new Panel(new FlowLayout(FlowLayout.CENTER,5,0));
  Panel b_p3=new Panel(new FlowLayout(FlowLayout.CENTER,5,0));
  Panel b_p4=new Panel(new FlowLayout(FlowLayout.CENTER,5,0));
  Panel bottom=new Panel();
  ScrollPane sp=new ScrollPane();
  CheckboxGroup treedrawer=new CheckboxGroup();
  Checkbox chk_simple=new Checkbox("Simple",treedrawer,true);
  Checkbox chk_detailed=new Checkbox("Detailed",treedrawer,false);
  
  /**
   * Initialize the applet.
   */
  public void init ()
  {
    super.init();
    String str;
  
    setLayout(new BorderLayout());
  
    t_loaddb.add(lbl_database);
    t_loaddb.add(tf_database);
    t_loaddb.add(btn_loaddb);
    t_p1.add(lbl_goal);
    t_p1.add(GoalAttributeChoice=new Choice());
    t_p2.add(lbl_entthres);
    t_p2.add(tf_entthres);
    t_p3.add(lbl_scorethres);
    t_p3.add(tf_scorethres);
    t_p4.add(lbl_learningset);
    t_p4.add(lbl_lssize);
    t_p4.add(tf_lssize);
    t_p4.add(lbl_lsmode);
    t_p4.add(SelectionChoiceLs=new Choice());
    t_p4.add(btn_starttree);
    SelectionChoiceLs.add("Random");
    SelectionChoiceLs.add("First");
    SelectionChoiceLs.add("Last");
    t_p5.add(lbl_candatt,"North");
    t_p5.add(lst_candatt=new java.awt.List(5,true),"Center");
    t_p0.add(t_p1) ; t_p0.add(t_p2) ; t_p0.add(t_p3);
    top.setLayout(new BorderLayout());
    top.add(t_loaddb, "North");
    top.add(t_p0, "West");
    top.add(t_p5,"Center");
    top.add(t_p4,"South");
    
    // center : the tree canvas
    sp.add(tc=new WalTreeCanvas());
   
    // bottom : control of the building process
    bottom.setLayout(new GridLayout(4,1,0,0));
    b_p1.add(tf_opennode=new TextField(100));
    tf_opennode.setEditable(false);
    bottom.add(b_p1);
    b_p3.add(AttChoice=new Choice());
    bottom.add(b_p3);
    b_p2.add(btn_scan=new Button("Scan attributes"));
    b_p2.add(btn_split=new Button("Split"));
    b_p2.add(btn_build=new Button("Build"));
    b_p2.add(chk_simple);
    b_p2.add(chk_detailed);
    b_p2.add(btn_refresh=new Button("Refresh"));
    b_p2.add(lbl_zoom);
    b_p2.add(btn_zoominc=new Button("+"));
    b_p2.add(btn_zoomdec=new Button("-"));
    bottom.add(b_p2);

    b_p4.add(lbl_testset);
    b_p4.add(lbl_tssize);
    b_p4.add(tf_tssize);
    b_p4.add(lbl_tsmode);
    b_p4.add(SelectionChoiceTs=new Choice());
    SelectionChoiceTs.add("Not in ls");
    SelectionChoiceTs.add("Random");
    SelectionChoiceTs.add("First");
    SelectionChoiceTs.add("Last");
    SelectionChoiceTs.add("Learning set");
    b_p4.add(btn_testtree=new Button("Test tree"));
    b_p4.add(btn_crossvalidation=new Button("Cross Validation"));
    b_p4.add(tf_testresult);
    tf_testresult.setEditable(false);
    bottom.add(b_p4);

    add(top, "North");
    add(bottom, "South");
    add(sp, "Center");

    snd=new SimpleNodeDrawer(tc);
    dnd=new DetailedNodeDrawer(tc);
    tc.nd=snd;

    tc.setParentDistance(30);

    // load the database
    if (getParameter("DBFILE")!=null) { 
      URL u;
      try {
	u=new URL(getDocumentBase(), getParameter("DBFILE"));
	initDb(u);
      } catch (MalformedURLException e) {
	System.out.println("-- database file not found");
	showStatus("Bad url");
      }
      
    }
    
    // different actions binding
    //   buttons
    btn_split.addActionListener(this);
    btn_refresh.addActionListener(this);
    btn_scan.addActionListener(this);
    btn_starttree.addActionListener(this);
    btn_testtree.addActionListener(this);
    btn_crossvalidation.addActionListener(this);
    btn_build.addActionListener(this);
    btn_loaddb.addActionListener(this);
    btn_zoominc.addActionListener(this);
    btn_zoomdec.addActionListener(this);
    //   check button
    chk_simple.addItemListener(this);
    chk_detailed.addItemListener(this);
    //   selection list
    SelectionChoiceTs.addItemListener(this);
    //   mouse
    MyMouseListener ml=new MyMouseListener();
    tc.addMouseListener(ml);
  }

  public void loadDb() {
    URL u;
    try {
      u=new URL(tf_database.getText());
      initDb(u);
      tc.setTree(root);
      tc.repaint();
    } catch (MalformedURLException e) {
      System.out.println("-- database file not found");
      showStatus("Bad url");
    }
  }

  public void initDb(URL u) {
    db = new Database();
    System.out.println("Database loading:");
    showStatus("Loading database");
    
    if (db.loadDb(u)==1) {
	showStatus("Loading failed : bad file format");
	return;
    }
    
    tf_database.setText(u.toString());

    if (db.objects.size==0) {
      System.out.println("no object");
    }

    //-- fill candidate attribute selection list
    lst_candatt.removeAll();
    GoalAttributeChoice.removeAll();

    int last_symb=0;
    for (int i=0; i<db.attributes.size(); i++) {
      Attribute att=(Attribute)db.attributes.elementAt(i);
      lst_candatt.add(att.describe());
      lst_candatt.select(i);
      if (att instanceof SymbAtt) {
	GoalAttributeChoice.add(att.toString());
	last_symb=i;
      }
    }
    GoalAttributeChoice.select(GoalAttributeChoice.getItemCount()-1);
    lst_candatt.deselect(last_symb);

    //-- initialise score and entropy threshold
    tf_entthres.setText("" + default_entthres);
    tf_scorethres.setText("" + default_scorethres);
    
    //-- initialise ls size and type
    tf_lssize.setText("" + (int)(2*db.objects.size/3));
    SelectionChoiceLs.select(0);
    tf_tssize.setText("" + (int)(db.objects.size/3));
    SelectionChoiceTs.select(0);
    AttChoice.add(messagechoice);
    AttChoice.select(0);

    //-- initialise the tree
    root=null;
    
    //-- garbage collecting after this (to remove the previous database)
    System.gc();
    
    showStatus("done");
    setComment("");
  }
  
  public void NewTree() {
    //-- learning set selection
    int lssize=GetLsSize();
    int lsmode=GetLsMode();
    if (lsmode==0)
      ls=db.objects.selectRandom(lssize);
    else if (lsmode==1)
      ls=db.objects.selectFirst(lssize);
    else if (lsmode==2)
      ls=db.objects.selectLast(lssize);
    
    //-- we build a trivial tree
    CurrentDt=new DecisionTree(db, ls, GetCandAtt(), GetGoalAttribute(), 
			       GetEntThres(), GetScoreThres());
    CurrentDtn=CurrentDt.root;
    root=CurrentDt.root.graphnode;

    //-- we initialise the graphical interface
    setComment(CurrentDt.getNodeDescription(CurrentDtn));
    AttChoice.removeAll();
    AttChoice.add(messagechoice);
    AttChoice.select(0);
    refreshcurrenttree();
  }
  
  /**
   *  some functions to get the parameters value
   */
  public Vector GetCandAtt() {
    int[] index=lst_candatt.getSelectedIndexes();
    if (index.length==0)
      return new Vector();
    Vector result=new Vector();
    for (int i=0; i<index.length; i++) {
      result.addElement(db.attributes.elementAt(index[i]));
    }
    return result;
  }

  public SymbAtt GetGoalAttribute() {
    return (SymbAtt)db.findAttribute(GoalAttributeChoice.getSelectedItem());
  }

  public double GetEntThres() {
    return (new Double(tf_entthres.getText())).doubleValue();
  }

  public double GetScoreThres() {
    return (new Double(tf_scorethres.getText())).doubleValue();
  }

  public int GetLsSize() {
    if (tf_lssize.getText()=="")
      return 0;
    else
      return Integer.parseInt(tf_lssize.getText());
  }

  public int GetLsMode() {
    return SelectionChoiceLs.getSelectedIndex();
  }

  public int GetTsSize() {
    if (tf_lssize.getText()=="")
      return 0;
    else
      return Integer.parseInt(tf_tssize.getText());
  }

  public int GetTsMode() {
    return SelectionChoiceTs.getSelectedIndex();
  }
  
  public void setComment(String t) {
    tf_opennode.setText(t);
  }
  
  /**
     * Start the main applet thread.
     */
  public void start ()
  {
    running = true;
    t = new Thread(this);
    t.start();
  }

  /**
     * Stop the applet.
     */
  public void stop ()
  {
    running = false;
  }

  /**
     * Main event loop.
     */
  public void run () { }

  /** 
   *  scan all the attributes at the top node for the best attribute.
   *  display the result into the choice list.
   */
  public void Scanatt() {
    
    if (CurrentDtn==null) 
      return;
    
    Vector choices=CurrentDt.scanNode(CurrentDtn);
    AttChoice.removeAll();
    for (int i=0; i<choices.size(); i++)
      AttChoice.add((String)choices.elementAt(i));
  }

  /**
   *  Split the current node
   *
   */
  public void SplitMore() {

    if (CurrentDtn==null) 
      return;
    
    if (CurrentDtn.type!="open") {
      // if the node is not open, we open it
      CurrentDtn.open();
    }
    
    if (AttChoice.getItem(0)==messagechoice) {
      //-- the user has not asked for attributes scanning
      CurrentDt.scanNode(CurrentDtn);
      changeSelectedNode(CurrentDt.splitNode(CurrentDtn,0));
    } else {
      //-- we take into account the user's choice
      changeSelectedNode(CurrentDt.splitNode(CurrentDtn,AttChoice.getSelectedIndex()));
    }
    refreshcurrenttree();
  }

  public void changeSelectedNode(DtNode dtn) {
    if (CurrentDtn!=null)
      CurrentDtn.graphnode.selected=false;
    if (AttChoice.getItem(0)!=messagechoice) {
      AttChoice.removeAll();
      AttChoice.add(messagechoice);
      AttChoice.select(0);
    }
    CurrentDtn=dtn;
    if (CurrentDtn!=null) {
      CurrentDtn.graphnode.selected=true;
      setComment(CurrentDt.getNodeDescription(CurrentDtn));
    } else {
      setComment("");
    }
  }

  public void buildalltree() {
    showStatus("Building the tree...");
    CurrentDt.buildDt();
    changeSelectedNode(null);
    refreshcurrenttree();
    CurrentDt.printDt();
    showStatus("done");
  }

  public void testtree() {
    //-- learning set selection
    showStatus("Testing the tree...");
    int tssize=GetTsSize();
    int tsmode=GetTsMode();
    if (tsmode==0)
      ts=db.objects.selectNotIn(ls);
    else if (tsmode==1)
      ts=db.objects.selectRandom(tssize);
    else if (tsmode==2)
      ts=db.objects.selectFirst(tssize);
    else if (tsmode==3)
      ts=db.objects.selectLast(tssize);
    else if (tsmode==4)
      ts=ls;
    String result="" + CurrentDt.testDt(ts)*100;
    if (result.length()<5)
      tf_testresult.setText("" + result + "%");
    else
      tf_testresult.setText("" + result.substring(0,5) + "%");
    showStatus("done");
  }

  public void CrossValidation() {
    showStatus("Cross validation...please wait");
    double result=CurrentDt.testDtCrossvalidation(GetTsSize());
    String sr="" + result*100;
    if (sr.length()<5)
      tf_testresult.setText(sr + "%");
    else
      tf_testresult.setText(sr.substring(0,5) + "%");
    showStatus("done");
  }

  public void actionPerformed(ActionEvent event) {
    String command=event.getActionCommand();

    if (command=="Split" && root!=null) {
      SplitMore();
    } else if (command=="Build" && root!=null) {
      buildalltree();
    } else if (command=="Scan attributes" && root!=null) {
      Scanatt();
    } else if (command=="Refresh" && root!=null) {
      refreshcurrenttree();
    } else if (command=="Test tree" && root!=null) {
      testtree();
    } else if (command=="New Tree") {
      NewTree();
    } else if (command=="Load Database") {
      loadDb();
    } else if (command=="Cross Validation" && root!=null) {
      CrossValidation();
    } else if (command=="+") {
      snd.increaseZoom();
      dnd.increaseZoom();
      tc.repaint();
    } else if (command=="-") {
      snd.decreaseZoom();
      dnd.decreaseZoom();
      tc.repaint();
    }
  }

  public void refreshcurrenttree() {
    if (root==null)
      return;
    tc.nd.computeNodeSize(root);
    tc.setTree(root);
    tc.repaint();
  }

  public void itemStateChanged(ItemEvent event) {
    String command=(String)event.getItem();
    
    if (command == "Simple") {
      tc.nd=snd;
      refreshcurrenttree();
    } else if (command == "Detailed") {
      tc.nd=dnd;
      if (root!=null) {
	dnd.initWidth(root);
	refreshcurrenttree();
      }
    } else if (command == "Not in ls") {
      tf_tssize.setText("" + (db.objects.size - Integer.parseInt(tf_lssize.getText())));
    } else if (command == "Learning set") {
      tf_tssize.setText(tf_lssize.getText());
    }
  }

  /* gestion de la souris */
  class MyMouseListener extends MouseAdapter {
    public void mousePressed(MouseEvent e) {
      int x=e.getX();
      int y=e.getY();
      if (root!=null) {
	DtNode dtn=root.getPointedNode((int)(x/tc.nd.zoom),(int)(y/tc.nd.zoom));
	if (dtn!=null) {
	  // select this node
	  changeSelectedNode(dtn);
	  // retrace the graph (there must exist a better way to do this)
	  tc.repaint();
	}
      }
    } 
  }
}
