/*
 *    ObjectSet.java
 *    Copyright (C) 1999 Pierre Geurts
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

import java.util.*;

/* simplement un vecteur d'object, on verra apres si on garde
   ce niveau d'abstraction ou pas
   */
   
public class ObjectSet {
  /* objects list*/
  Vector objectslist;
  /* a cache value for the entropy in this set */
  double entropy;
  /* the size of the learning set */
  int size;
  /* summary, the frequence of each class in this set */
  int summary[];
  
  ObjectSet(Vector vo) {
    objectslist=vo;
    size=vo.size();
  }

  ObjectSet() {
    objectslist=new Vector();
    size=0;
  }

/* only for compatibility (primary an objectset was simply a vector) */
  public int size() {
    return size;
  }

  public Obj elementAt(int index) {
    return (Obj)objectslist.elementAt(index);
  }
  public void addElement(Obj o) {
    objectslist.addElement(o);
    size++;
  }
  public void setElementAt(Obj o, int index) {
    objectslist.setElementAt(o, index);
  }

  public int indexOf(Obj o) {
    return objectslist.indexOf(o);
  }

  public String toString() {
    return objectslist.toString();
  }

/* split a set according to the given test */
  public Vector splitOn(Test t) {
    int index;
    Vector v=new Vector();
    Obj o;
    for (int i=0; i<t.nbIssues; i++) {
      v.addElement(new ObjectSet());
    }
    for (int i=0; i<size; i++) {
      o=(Obj)objectslist.elementAt(i);
      index=t.TestIndex(o);
      ((ObjectSet)v.elementAt(index)).addElement(o);
    }
    return v;
  }
/* methods to select subset of objects from the database */

/* select the first n objects from this subset */
  public ObjectSet selectFirst(int n) {
    ObjectSet os=new ObjectSet();
    for (int i=0;i<n && i<size;i++) {
      os.addElement((Obj)objectslist.elementAt(i));
    }
    return os;
  }

/* select the last n objects from this subset */
  public ObjectSet selectLast(int n) {
    ObjectSet os=new ObjectSet();
    int i;
    if (n>size)
      i=0;
    else 
      i=size-n;
    for (;i<size;i++) {
      os.addElement((Obj)objectslist.elementAt(i));
    }
    return os;
  }

/* select randomly n objects from this set (very stupid) */
  public ObjectSet selectRandom(int n) {
    ObjectSet os=new ObjectSet();
    Vector v=(Vector)objectslist.clone();
    int sv=size;
    int pos;
    if (n>size) 
      n=size;
    for (int i=0;i<n;i++) {
      pos=(int)(sv*java.lang.Math.random());
      os.addElement((Obj)v.elementAt(pos));
      v.removeElementAt(pos);
      sv--;
    }
    return os;
  }

/* select objects which are not in a given subset */
  public ObjectSet selectNotIn(ObjectSet oset) {
    ObjectSet os=new ObjectSet();
    Obj o;
    for (int i=0;i<size;i++) {
      o=(Obj)objectslist.elementAt(i);
      if (oset.indexOf(o)==-1) 
	os.addElement(o);
    }
    return os;
  }
   
/* method to compute the entropy of an symbolic attribute in this subset
 * it computes also the subset summary
 */
  public double computeEntropy(SymbAtt goal) {
    int sum=size;
    int i,x,gvs=goal.values.size();
    summary=new int[gvs];
    for (i=0; i<sum; i++) {
      x=goal.values.indexOf(goal.Value((Obj)objectslist.elementAt(i)));
      summary[x]++;
    }
    entropy=Entropy.entropy(summary, gvs, sum);
    return entropy;
  }

/* give a textual description of the subset summary */
  public String getSummaryString(SymbAtt goal) {
    String s="";
    for (int i=0; i<summary.length; i++) {
      s+=(String)goal.values.elementAt(i);
      s+=":" + summary[i]+ " ";
    }
    s+="total:" + size;
    return s;
  }

/* method that gives the majority class in this subset (using the summary) */
  public String getMajClass(SymbAtt goal) {
    int imax=0;
    int maxfreq=summary[0];
    for (int i=1; i<summary.length; i++) {
      if (summary[i]>maxfreq) {
	maxfreq=summary[i];
	imax=i;
      }
    }
    return (String)goal.values.elementAt(imax);
  }

/* order a subset on numerical attribute values 
 * use quicksort algorithm
 */
  public void orderOn(NumAtt att) {
    orderOn(att,0,size-1);
  }

  public void orderOn(NumAtt att, int lo0, int hi0) {
    int lo=lo0;
    int hi=hi0;
    float mid;

    if (hi0>lo0) {
      mid = ((Float)att.Value((Obj)objectslist.elementAt((lo0+hi0)/2))).floatValue();
      while (lo<= hi) {
	while((lo<hi0) && 
	      (((Float)att.Value((Obj)objectslist.elementAt(lo))).floatValue()<mid))
	  ++lo;
	while((hi>lo0) && 
	      (((Float)att.Value((Obj)objectslist.elementAt(hi))).floatValue()>mid))
	  --hi;
	if(lo<=hi) {
	  Object tmp=objectslist.elementAt(lo);
	  objectslist.setElementAt(objectslist.elementAt(hi),lo);
	  objectslist.setElementAt(tmp,hi);
	  ++lo;
	  --hi;
	}
      }
      if(lo0<hi)
	orderOn(att,lo0,hi);
      if (lo<hi0)
	orderOn(att,lo,hi0);
    }
  }
}
