/*
 *    SimpleNodeDrawer.java
 *    Copyright (C) 1999 Pierre Geurts
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

import java.awt.*;
import java.util.*;

/**
 *  a abstract class to allow many representation for
 *  a node
 */

public class SimpleNodeDrawer extends NodeDrawer {
  
  FontMetrics metrics_small,metrics_big,metrics_current;
  Font smallfont, bigfont;
  int SMALL_FIXED_FONT_HEIGHT=8;
  int BIG_FIXED_FONT_HEIGHT=14;
  int current_font_height;
  int default_border=5;
  Vector colours=new Vector();
  Canvas cv;

  SimpleNodeDrawer(Canvas c) {
    cv=c;
    bigfont=new Font("Helvetica", Font.PLAIN, BIG_FIXED_FONT_HEIGHT);
    c.setFont(bigfont);
    metrics_big=c.getFontMetrics(c.getFont());
    
    smallfont=new Font("Helvetica", Font.PLAIN, SMALL_FIXED_FONT_HEIGHT);
    c.setFont(smallfont);
    metrics_small=c.getFontMetrics(c.getFont());

    metrics_current=metrics_small;
    current_font_height=SMALL_FIXED_FONT_HEIGHT;

    colours.addElement(Color.black);
    colours.addElement(Color.red);
    colours.addElement(Color.green);
    colours.addElement(Color.blue);
    colours.addElement(Color.orange);
    colours.addElement(Color.darkGray);
    colours.addElement(Color.cyan);
    colours.addElement(Color.gray);
    colours.addElement(Color.magenta);
    colours.addElement(Color.pink);
    colours.addElement(Color.black);
    colours.addElement(Color.yellow);
    colours.addElement(Color.white);
    colours.addElement(Color.red);
    colours.addElement(Color.green);
    colours.addElement(Color.blue);
    colours.addElement(Color.orange);
    colours.addElement(Color.darkGray);
    colours.addElement(Color.cyan);
    colours.addElement(Color.gray);
    colours.addElement(Color.magenta);
    colours.addElement(Color.pink);
    colours.addElement(Color.black);
    colours.addElement(Color.yellow);
    colours.addElement(Color.white);
  }
  
  public void computeNodeSize(WNode t) {
    t.width=2 * current_font_height;
    t.height=metrics_current.stringWidth(t.label) + 10;
    t.border=default_border;
    
    if (t.sibling !=null) {
      computeNodeSize(t.sibling);
    }
    
    if (t.child != null) {
      computeNodeSize(t.child);
    }
  }
  
  public void drawArc(Graphics g, int x1, int y1, int x2, 
		      int y2, String arclabel) {
    g.setColor(Color.black);
    g.drawLine((int)(x1*zoom), (int)(y1*zoom), (int)(x2*zoom),(int)(y2*zoom));
    g.drawString(arclabel,
		 (int)(((x1 + x2)*zoom - metrics_current.stringWidth(arclabel)) / 2),
		 (int)((y1 + y2)*zoom/2));
  }

  public void drawNode(Graphics g, WNode t) {
    int tposx=(int)(t.pos.x*zoom);
    int tposy=(int)(t.pos.y*zoom);
    int tw=(int)(t.width*zoom);
    int th=(int)(t.height*zoom);

    // Draw highlights
    g.setColor(Color.lightGray);
    g.drawLine(tposy + 2, tposx + tw + 1, tposy + th, 
	       tposx + tw + 1);
    g.drawLine(tposy + th + 1, tposx + tw + 1, 
	       tposy + th + 1, tposx + 2);

    // we draw the box corresponding to this node
    if (t.colour>=colours.size())
      g.setColor(Color.black);
    else
      g.setColor((Color)colours.elementAt(t.colour));

    g.drawRect(tposy, tposx, th, tw);
    
    // we draw the node label
    g.drawString(t.label, 
		 tposy + (th - metrics_current.stringWidth(t.label)) / 2,
		 tposx + tw - (tw - current_font_height) / 2);

    if (t.selected) {
      int border=(int)(3*zoom);
      g.setColor(Color.pink);
      g.drawRect(tposy-border, tposx-border, th+2*border, tw+2*border);
    }
    
  }
  
  // increase the zoom
  public void increaseZoom() {
    if (zoom==0.5)
      zoom=0.7;
    else if (zoom==0.7)
      zoom=1.0;
    else if (zoom==1.0)
      zoom=1.2;
    else if (zoom==1.2) {
      zoom=1.5;
      metrics_current=metrics_big;
      current_font_height=BIG_FIXED_FONT_HEIGHT;
      cv.setFont(bigfont);
    } else if (zoom==1.5)
      zoom=2.0;
    else if (zoom==2.0)
      zoom=2.5;
    else if (zoom==2.5)
      zoom=3.0;
  }
  
  // decrease the zoom
  public void decreaseZoom() {
    if (zoom==3.0)
      zoom=2.5;
    else if (zoom==2.5)
      zoom=2.0;
    else if (zoom==2.0)
      zoom=1.5;
    else if (zoom==1.5) {
      zoom=1.2;
      metrics_current=metrics_small;
      current_font_height=SMALL_FIXED_FONT_HEIGHT;
      cv.setFont(smallfont);
    } else if (zoom==1.2)
      zoom=1.0;
    else if (zoom==1.0)
      zoom=0.7;
    else if (zoom==0.7)
      zoom=0.5;
  }
}
