/*
 *    SymbAtt.java
 *    Copyright (C) 1999 Pierre Geurts
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* recherche de la meilleure partition binaire de l'ensemble des valeurs de
   l'attribut */

import java.util.*;
import java.io.*;

public class SymbAtt extends Attribute{
    Vector values=new Vector();
    /* la valeur de l'attribut symbolique qui donne le meilleur score */
    Vector bestsubset;

    SymbAtt(String s, Database dab) {
	name=s;
	db=dab;
	index=db.addAttToDb(this);
    }
    public void SetAttValue(Obj o,Object s) {
	int i;
	for (i=0; i<values.size(); i++)
	    if (s.equals(values.elementAt(i)))
		break;
	if (i<values.size())
	    o.attributesvalues.setElementAt(values.elementAt(i),index);
	else {
	    o.attributesvalues.setElementAt(s,index);
	    values.addElement(s);
	}
    }
    public int ReadAttValue(Obj o, StreamTokenizer stt) {
	if (stt.ttype==stt.TT_WORD) {
	    SetAttValue(o,stt.sval);
	    return 0;
	} else if (stt.ttype==stt.TT_NUMBER) {
	    // we transform number to a string
	    SetAttValue(o,"" + stt.nval);
	    return 0;
	} else {
	    System.out.println("A symbolic value was expected");
	    return 1;
	}
    }  

    public String describe() {
	return toString()+ " - symbolic - values : " + values.toString();
    }

    public String describeBestTest() {
	return toString() + " - values : " + bestsubset.toString() + " - Information gain : " + lastbestscore;
    }
  
    public double bestscore(ObjectSet os, SymbAtt goal) {
	/* attention cette partie est tres tres mal ecrite */
	
	double result;
	double tmpresult;
	double currentbestscore=-1;
	int currentbestvalue=0;
	int pbv;
	int i,j,x,y;

	/* liste des index*/
	Vector remainingvalues=new Vector();
	for (i=0; i<values.size(); i++) {
	    remainingvalues.addElement(new Integer(i));
	}

	/* les valeurs dans l'ordre dans lequel elles sont selectionnees */
	Vector ordervalues=new Vector();
	/* les valeurs de score */
	Vector scores=new Vector();

	int gvs=goal.values.size();
	int sum=os.size;
	    
	/* pour chaque valeur de l'attribut, on conserve la matrice des frequences */
	int tabfreq[][][]=new int[values.size()][2][gvs+1];
	
	/* calcul des matrices de frequences */
	for (j=0; j<values.size(); j++) {
	    
	    for (i=0; i<sum; i++) {
		x=values.indexOf(this.Value((Obj)os.elementAt(i)));
		y=goal.values.indexOf(goal.Value((Obj)os.elementAt(i)));
		
		if (x==j) { /* la valeur est la valeur testee */
		    tabfreq[j][0][y]++;
		    tabfreq[j][0][gvs]++;
		} else {
		    tabfreq[j][1][y]++;
		    tabfreq[j][1][gvs]++;
		}
	    }
	    /* calcul de l'entropie */
	    result=os.entropy;
	    result-=tabfreq[j][0][gvs]*Entropy.entropy(tabfreq[j][0], gvs,tabfreq[j][0][gvs])/sum;
	    result-=tabfreq[j][1][gvs]*Entropy.entropy(tabfreq[j][1], gvs,tabfreq[j][1][gvs])/sum;
	    
	    if (result>currentbestscore) {
		currentbestscore=result;
		currentbestvalue=j;
	    }
	}
	
	/* on va maintenant regrouper les attributs de maniere recursive */
	int currenttabfreq[][]=new int[2][gvs+1];
	for (i=0;i<gvs+1;i++) {
	    currenttabfreq[0][i]=0;
	    currenttabfreq[1][i]=tabfreq[currentbestvalue][0][i]
		+ tabfreq[currentbestvalue][1][i];
	}
	//	printtab(currenttabfreq,gvs+1);

	/* on ajoute cet element dans la liste */
	Integer cbv=new Integer(currentbestvalue);
	ordervalues.addElement(cbv);
	scores.addElement(new Double(currentbestscore));
	i=0;
	while (i<remainingvalues.size() &&
	       ((Integer)remainingvalues.elementAt(i)).intValue()!=currentbestvalue)
	    i++;
	remainingvalues.removeElementAt(i);

	while (remainingvalues.size()!=1) {
	    /* copier le tableau courant dans la valeur correspondant
	       a l'element precedent */
	    for (i=0;i<gvs+1;i++) {
		currenttabfreq[0][i]+=tabfreq[currentbestvalue][0][i];
		currenttabfreq[1][i]-=tabfreq[currentbestvalue][0][i];
	    }
	    //	    printtab(currenttabfreq,gvs+1);
	    /* on va utiliser la matrice correspondant a la precedente
	       valeur comme moyen de stockage temporaire
	       */

	    pbv=currentbestvalue;

	    /* recherche de la meilleure valeur a ajouter au sous-ensemble */
	    currentbestscore=-1;
	    
	    for (j=0; j<remainingvalues.size(); j++) {
		int cv=((Integer)remainingvalues.elementAt(j)).intValue();
		for (i=0; i<gvs+1; i++) {
		    tabfreq[pbv][0][i]=tabfreq[cv][0][i]+currenttabfreq[0][i];
		    tabfreq[pbv][1][i]=currenttabfreq[1][i]-tabfreq[cv][0][i];
		}
		
		result=os.entropy;
		result-=tabfreq[pbv][0][gvs]*
		    Entropy.entropy(tabfreq[pbv][0], gvs, tabfreq[pbv][0][gvs])/sum;
		result-=tabfreq[pbv][1][gvs]*
		    Entropy.entropy(tabfreq[pbv][1], gvs, tabfreq[pbv][1][gvs])/sum;
		if (result>currentbestscore) {
		    currentbestscore=result;
		    currentbestvalue=cv;
		}
	    }
	    cbv=new Integer(currentbestvalue);
	    ordervalues.addElement(cbv);
	    scores.addElement(new Double(currentbestscore));
	    i=0;
	    while (i<remainingvalues.size() &&
		   ((Integer)remainingvalues.elementAt(i)).intValue()!=currentbestvalue)
		i++;
	    remainingvalues.removeElementAt(i);
	}
	
	/* recherche du meilleur score */
	currentbestscore=((Double)scores.elementAt(0)).doubleValue();
	currentbestvalue=0;
	for (i=1; i<ordervalues.size(); i++) {
	    double sc=((Double)scores.elementAt(i)).doubleValue();
	    if (sc>currentbestscore) {
		currentbestscore=sc;
		currentbestvalue=i;
	    }
	}

	/* on garde uniquement les elements de l'ensemble */
	ordervalues.setSize(currentbestvalue+1);
	bestsubset=new Vector();
	for (i=0; i<ordervalues.size(); i++)
	    bestsubset.addElement(values.elementAt(((Integer)ordervalues.elementAt(i)).intValue()));
	
	//	System.out.println(name + " " + values.toString() + " " + bestsubset.toString());
	
	lastbestscore=currentbestscore;
	return currentbestscore;
    }

    public Partition bestpartition(ObjectSet os, SymbAtt goal) {
	TestSymb t=new TestSymb(this, bestsubset);
	Vector v=os.splitOn(t);
	return new Partition(t,v);
    }

    /*    
    public static void printtab(int[][] tab,int size) {
	System.out.print("[["); 
	for (int i=0; i<size; i++) {
	    System.out.print("" + tab[0][i] + ", ");  
	}
	System.out.println("]");
	System.out.print("[");
	for (int i=0; i<size; i++) {
	    System.out.print("" + tab[1][i] + ", ");  
	}
	System.out.println("]]");
    }
    */
	
}
