import java.awt.Point;

/**
 * Represents a node (or tree) in the WalTree.
 */
public class WNode {
  String  label;         // the label of the node (test)
  String  arclabel;      // the label of the arc leading to this node
  WNode   parent, child, sibling;
  int     width, height, border; // the width, height and border of this node
  Point   pos, offset;
  Polygon contour;      
  int colour=0;  // a colour number
  DtNode dtnode; // the associated DtNode
  boolean selected=false; // tell us if this node is the current node
  
  public WNode(String l, String a, WNode p, WNode c, WNode s, int w, int h, int b, DtNode dtn) {
    label = l;
    arclabel=a;
    parent = p;
    child = c;
    sibling = s;
    width = w;
    height = h;
    border = b;
    pos = new Point(0, 0);
    offset = new Point(0, 0);
    contour = new Polygon();
    dtnode=dtn;
  }

  public boolean containsPoint(int x, int y) {
    return (((y>=pos.x) && (y<=pos.x+width)) && ((x>=pos.y) && (x<=pos.y+height)));
  }
  
  public DtNode getPointedNode(int x,int y) {
    if (containsPoint(x,y))
      return dtnode;
    if (child!=null && y>pos.x) { 
      // not worth going further down if y<pos.x
      DtNode dtchild=child.getPointedNode(x,y);
      if (dtchild!=null)
	return dtchild;
    }
    if (sibling!=null)
      return sibling.getPointedNode(x,y);
    return null;
  }
}
